import struct
from typing import Union, TextIO
from word_vectors import Vocab, Vectors
from file_or_name import file_or_name


def _pad(word: str, max_len: int) -> bytes:
    """Pad a word out so the byte representation is max_len."""
    b = word.encode("utf-8")
    return b + b" " * (max_len - len(b))


@file_or_name(wf="wb")
def write_dense(wf: Union[str, TextIO], vocab: Vocab, vectors: Vectors, max_len: int) -> None:
    """Write vectors to a dense file.

    Args:
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
        max_len: The longest length of the words as bytes.
        file_name: Where to save the file.
    """
    for val in (len(vocab), vectors.shape[1], max_len):
        wf.write(struct.pack("<i", val))
    for word, vector in zip(vocab, vectors):
        word = _pad(word, max_len)
        wf.write(word)
        wf.write(vector.tobytes())


@file_or_name(wf="w")
def write_glove(wf: Union[str, TextIO], vocab: Vocab, vectors: Vectors) -> None:
    """Write vectors to a dense file.

    Args:
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
        file_name: Where to save the file.
    """
    for word, vector in zip(vocab, vectors):
        wf.write(" ".join([word, *map(str, vector)]) + "\n")


@file_or_name(wf="wb")
def write_w2v(wf: Union[str, TextIO], vocab: Vocab, vectors: Vectors) -> None:
    """Write vectors to a dense file.

    Args:
        vocab: The vocab of words -> ints.
        vectors: The vectors as a np.ndarray.
        file_name: Where to save the file.
    """
    wf.write(f"{len(vocab)} {vectors.shape[1]}\n".encode("utf-8"))
    for word, vector in zip(vocab, vectors):
        wf.write((word + " ").encode("utf-8") + vector.tobytes())



__version__ = "1.0.2"

from typing import Dict, Tuple
import numpy as np

Vocab = Dict[str, int]
Vectors = np.ndarray

INT_SIZE = 4
FLOAT_SIZE = 4
DENSE_HEADER = 3

import word_vectors.read as read_module
from word_vectors.read import read, read_w2v, read_glove, read_dense, FileType
from word_vectors.write import write_w2v, write_glove, write_dense



import re
import os
import mmap
import struct
import pathlib
from enum import Enum
from typing import Iterable, Tuple, Union, TextIO
import numpy as np
from file_or_name import file_or_name
from word_vectors import INT_SIZE, FLOAT_SIZE, DENSE_HEADER, Vocab, Vectors
from word_vectors.write import write_dense

FileType = Enum("FileType", "GLOVE W2V DENSE")

glove = re.compile(br"^[^ ]+? (-?\d+?\.\d+? )+", re.MULTILINE)
w2v = re.compile(br"^\d+? \d+?$", re.MULTILINE)


@file_or_name(f="rb")
def sniff(f: Union[str, TextIO]) -> FileType:
    """Figure out what kind of vector file it is."""
    b = f.read(1024)
    if f.mode == "r":
        return FileType.GLOVE
    if w2v.match(b):
        return FileType.W2V
    elif glove.match(b):
        return FileType.GLOVE
    return FileType.DENSE


def find_max(words: Iterable[str]) -> int:
    """Get the max length of words (as bytes)."""
    return max(map(len, map(lambda x: x.encode("utf-8"), words)))


def read(f: Union[str, TextIO], convert: bool = False, replace: bool = False) -> Tuple[Vocab, Vectors]:
    """Read vectors from file.

    Args:
        f: The file to read from.
        convert: Convert the vectors into the Dense format.
        replace: Replace the vector file with the Dense version.

    Returns:
        The vocab and vectors.
    """
    w = None
    wv = None
    len_ = None
    type_ = sniff(f)
    if type_ is FileType.GLOVE:
        w, wv = read_glove(f)
    elif type_ is FileType.W2V:
        w, wv, len_ = read_w2v(f, stats=True)
    elif type_ is FileType.DENSE:
        w, wv, len_ = read_dense(f, stats=True)
        convert = False
        replace = False
    if convert:
        if isinstance(f, (str, pathlib.PurePath)):
            output = str(f)
        else:
            output = f.name
        if not replace:
            output = output + ".dense"
        if len_ is None:
            len_ = find_max(w)
        write_dense(output, w, wv, len_)
    return w, wv


@file_or_name(f="r")
def read_glove(f: str) -> Tuple[Vocab, Vectors]:
    """Read vectors from a glove file.

    Args:
        file_name: The file to read from

    Returns:
        The vocab and vectors.
    """
    words = {}
    vectors = []
    i = 0
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        for line in iter(m.readline, b""):
            line = line.decode("utf-8")
            line = line.rstrip("\n")
            word, *vector = line.split(" ")
            if word not in words:
                words[word] = i
                vectors.append(np.asarray(vector, dtype=np.float32))
                i += 1
    return words, np.vstack(vectors)


def _find_space(m, offset) -> Tuple[str, int, int]:
    i = offset + 1
    while m[i : i + 1] != b" ":
        i += 1
    word = m[offset:i].decode("utf-8")
    return word, i - offset, i + 1


@file_or_name(f="rb")
def read_w2v(f: str, stats: bool = False) -> Tuple[Vocab, Vectors]:
    """Read vectors from a word2vec file.

    Args:
        file_name: The file to read from
        stats: Should you return the length of the longest word.

    Returns:
        The vocab and vectors.
    """
    words = {}
    vectors = []
    max_len = 0
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        header = m.readline()
        offset = m.tell()
        vocab, dim = map(int, header.decode("utf-8").split())
        size = FLOAT_SIZE * dim
        for i in range(vocab):
            word, word_len, offset = _find_space(m, offset)
            if word_len > max_len:
                max_len = word_len
            raw = m[offset : offset + size]
            vector = np.frombuffer(raw, dtype=np.float32)
            words[word] = i
            vectors.append(vector)
            offset = offset + size
    if stats:
        return words, np.vstack(vectors), max_len
    return words, np.vstack(vectors)


@file_or_name(f="rb")
def read_dense(f: str, stats: bool = False) -> Tuple[Vocab, Vectors]:
    """Read vectors from a dense file.

    Args:
        file_name: The file to read from
        stats: Should you return the length of the max word?

    Note:
        The dense file format is pretty simple. The first three bytes are
        the vocab size, vector size, and max word length as little endian
        values. Each "line" is a word in max word bytes followed by the
        vector of size vector size * 4 (size of np.float32)

    Returns:
        The vocab and vectors.
    """
    words = {}
    vectors = []
    with mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ) as m:
        offset = INT_SIZE * DENSE_HEADER
        vocab, dim, length = struct.unpack("<iii", m[:offset])
        size = FLOAT_SIZE * dim
        for i in range(vocab):
            start = offset + i * (length + size)
            line = m[start : start + length + size]
            word = line[:length].decode("utf-8").strip(" ")
            vector = np.frombuffer(line[length:], dtype=np.float32)
            words[word] = i
            vectors.append(vector)
    if stats:
        return words, np.vstack(vectors), length
    return words, np.vstack(vectors)



import ast
from typing import Optional
from setuptools import setup, find_packages


def get_version(file_name: str, version_name: str = "__version__") -> Optional[str]:
    with open(file_name) as f:
        tree = ast.parse(f.read())
        for node in ast.walk(tree):
            if isinstance(node, ast.Assign):
                if node.targets[0].id == version_name:
                    return node.value.s
    raise ValueError(f"Couldn't find assignment to variable {version_name} in file {file_name}")


class About(object):
    NAME = "word-vectors"
    VERSION = get_version("word_vectors/__init__.py")
    AUTHOR = "blester125"
    EMAIL = f"{AUTHOR}@gmail.com"
    URL = f"https://github.com/{AUTHOR}/{NAME}"
    DL_URL = f"{URL}/archive/{VERSION}.tar.gz"
    LICENSE = "MIT"
    DESCRIPTION = "Word Vectors"


ext_modules = []


setup(
    name=About.NAME,
    version=About.VERSION,
    description=About.DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author=About.AUTHOR,
    author_email=About.EMAIL,
    url=About.URL,
    download_url=About.DL_URL,
    license=About.LICENSE,
    python_requires=">=3.6",
    packages=find_packages(),
    package_data={"word_vectors": [],},
    include_package_data=True,
    install_requires=["file_or_name>=1.1.3", "numpy"],
    extras_require={"test": ["pytest"],},
    keywords=[],
    ext_modules=ext_modules,
    entry_points={"console_scripts": [],},
    classifiers={
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    },
)
